#!/usr/bin/perl

use Text::MediawikiFormat 'wikiformat';
use DBI;
use Digest::MD5 qw(md5_hex);
use File::Copy;
use File::MMagic;
use strict;
use vars qw($txtId %pages %namespaces $table_prefix $is_clean);

###################################################
# CONFIGURATION SECTION
###################################################

# Namespaces to process
%namespaces = (0 => 'Main',
                100 => 'UXD',
                200 => 'PM',
                320 => 'SearchCenter',
                330 => 'Emerging_Tech');


# MediaWiki DB settings
my $mw_dbname = 'dbname';
my $mw_dbuser = 'dbuser';
my $mw_dbpass = 'dbpass';
$table_prefix = 'pd_';

###################################################

$is_clean = 1;
my $mw_dbh = DBI->connect("DBI:mysql:$mw_dbname", $mw_dbuser, $mw_dbpass) or die "Couldn't connect to database: " . DBI->errstr;

# Get page IDs of pages belonging  to @namespaces 
my $stsel = $mw_dbh->prepare("SELECT page_id FROM ${table_prefix}page WHERE page_namespace = ?") or die "Couldn't prepare SELECT statement: " . $mw_dbh->errstr;
# Get text ID of latest revision for each page
my $stsel1 = $mw_dbh->prepare("SELECT rev_text_id FROM ${table_prefix}revision WHERE rev_id = (SELECT page_latest FROM ${table_prefix}page WHERE page_id = ?)") or die "Couldn't prepare SELECT statement: " . $mw_dbh->errstr;

my ($row, $subrow);

foreach (keys %namespaces){

    $stsel->bind_param(1, $_);
    $stsel->execute or die "Couldn't execute SELECT statement: " . $stsel->errstr;

    while ($row = $stsel->fetchrow_hashref){
    
	$stsel1->bind_param(1, $row->{page_id});
	$stsel1->execute or die "Couldn't execute SELECT statement: " . $stsel1->errstr;
	
	while ($subrow = $stsel1->fetchrow_hashref){
		
	    $pages{$subrow->{rev_text_id}} = $row->{page_id};
	    
	}
    }
}


$stsel = $mw_dbh->prepare("SELECT old_text FROM ${table_prefix}text WHERE old_id = ?") or die "Couldn't prepare SELECT statement: " . $mw_dbh->errstr;
$stsel1 = $mw_dbh->prepare("SELECT page_title, page_namespace FROM ${table_prefix}page WHERE page_id = ?") or die "Couldn't prepare SELECT statement: " . $mw_dbh->errstr;

my $text;        
my $pname;

foreach $txtId (keys %pages){

    $pname = '';

    $stsel1->bind_param(1, $pages{$txtId});
    $stsel1->execute or die "Couldn't execute SELECT statement: " . $stsel1->errstr;


    while ($subrow = $stsel1->fetchrow_hashref){
		
        $pname .= "$namespaces{$subrow->{page_namespace}}:" if $subrow->{page_namespace} != 0;
        $pname .= $subrow->{page_title};
    
    
    }

    $stsel->bind_param(1, $txtId);
    $stsel->execute or die "Couldn't execute SELECT statement: " . $stsel->errstr;


    while ($row = $stsel->fetchrow_hashref()){

	$text = $row->{old_text};
	

    }

    $text = checkTables($text, $pname) if ( $text =~ /\{\|/m );

     if ( $text =~ /\{\|/m ){

	print "MediaWiki page '$pname' contains table(s) without closing tag '|}'!\n";
        $is_clean = 0;
    }

    
    
}   

if ($is_clean){

    print "\nNo issues detected! Proceed with data conversion.\n\n";

}else{
    
    print "\nPlease fix issues on pages listed above before you start data conversion!\n\n";
}

$mw_dbh->disconnect;




sub checkTables{


    my $txt = shift;
    my $pnam = shift;
    my $sstr;


    while($txt =~ m/(\{\|.+?\|\})/s){

        $sstr = $1;

        if ($sstr =~ m/^\{\|.+\{\|.+\|\}/s){

            $txt =~ s/\{\|(.+?\|\})/eval{my $tt = checkTables($1, $pnam); return ('{|'. $tt);}/es;

        }else{

            $txt =~ s/(\{\|.+?\|\})/checkTable($1, $pnam)/es;

        }
    }

    return $txt;
}

sub checkTable{

    my $tline = shift;
    my $pn = shift;
    my (%table, $row, $cell, $tmp, $j);
    my $i = 0; # Row index


    # Parse the wiki table

    $tline =~ s/\|\}$//m;
    $tline =~ s/^\s+//;

    foreach $row (split /\|-/, $tline){

        $j = -1; # Cell index

        # Remove leading spaces
        $row =~ s/^\s+//;
        # Remove leading newline characters
        $row =~ s/^\n+//;


        # If this is the first row in the table
        if ($row =~/^\{\|(.*)/){

            $row =~ s/^\{\|.*//; # Cut it off
            $row =~ s/^\n+//;
        }

        # We got row with its own set of attributes
        if ($row =~ /^([^\|!].*)/){

            $row =~ s/.+//;
            $row =~ s/^\n+//;
        }


        # We got table with caption
        if ($row =~ /^\|\+(.+)/s){

            next;
        }

        # We got table in format "one row on one line"
        if ($row =~ m/^\|.+(?:\|\|.+)+/){

            next;

        }
        # The same as above but with table headings, not ordinary cells
        if($row =~ /^!.+(?:!!.+)+/){

           next;
        }

        # If we got here, we have table with one cell per line (row consits of multiple lines)

        foreach $cell (split /\n/, $row){

            if ($cell =~ /^\s*!(.*)/){ # Wow, <th>!

                $j++;

            }elsif ($cell =~ /^\s*\|(.*)/){ # <td>


                $j++;

            }elsif ($cell =~ /^([^\|!].*)/){ # Cell with multiline value

		if ($j == -1){

		    print "MediaWiki page '$pn' contains incorrect table markup! Some data can be lost during conversion!\n";
		    $is_clean = 0;
		    return;
#		    next;
		}

            }

        }

    }

    return;
}


