#!/usr/bin/perl

use Text::MediawikiFormat 'wikiformat';
use DBI;
use Digest::MD5 qw(md5_hex);
use File::Copy;
use File::MMagic;
use strict;
use vars qw($txtId %pages %namespaces $mw_att_base $tiki_att @attachments $table_prefix);

###################################################
# CONFIGURATION SECTION
###################################################

# Namespaces to process
%namespaces = (0 => 'Main',
                100 => 'UXD',
                200 =>'PM',
                320 => 'SearchCenter',
                330 =>'Emerging_Tech');

# MediaWiki base image directory (full path)
$mw_att_base = '/var/www/html/mediawiki/upload';

# TikiWiki attachments directory (full path)
$tiki_att = '/var/www/html/migration/files';

# MediaWiki DB settings
my $mw_dbname = 'dbname';
my $mw_dbuser = 'dbuser';
my $mw_dbpass = 'dbpass';
$table_prefix = 'pd_';

# Tiki DB settings

my $tiki_dbhost = 'localhost';
my $tiki_dbport = '3306';
my $tiki_dbname = 'dbname';
my $tiki_dbuser = 'dbuser';
my $tiki_dbpass = 'dbpass';
###################################################

# Get list of MW images and files
@attachments = listAttachments();


my $mw_dbh = DBI->connect("DBI:mysql:$mw_dbname", $mw_dbuser, $mw_dbpass) or die "Couldn't connect to database: " . DBI->errstr;
my $tiki_dbh = DBI->connect("DBI:mysql:database=$tiki_dbname;host=$tiki_dbhost;port=$tiki_dbport", $tiki_dbuser, $tiki_dbpass) or die "Couldn't connect to database: " . DBI->errstr;

# Recreate table for text converted to html

$mw_dbh->do("DROP TABLE IF EXISTS ${table_prefix}html") or die "Couldn't drop table: " . $mw_dbh->errstr;
$mw_dbh->do("CREATE TABLE IF NOT EXISTS ${table_prefix}html LIKE ${table_prefix}text") or die "Couldn't create table: " . $mw_dbh->errstr;

# Get page IDs of pages belonging  to @namespaces 
my $stsel = $mw_dbh->prepare("SELECT page_id FROM ${table_prefix}page WHERE page_namespace = ?") or die "Couldn't prepare SELECT statement: " . $mw_dbh->errstr;
# Get text ID of latest revision for each page
my $stsel1 = $mw_dbh->prepare("SELECT rev_text_id FROM ${table_prefix}revision WHERE rev_id = (SELECT page_latest FROM ${table_prefix}page WHERE page_id = ?)") or die "Couldn't prepare SELECT statement: " . $mw_dbh->errstr;

my ($row, $subrow);

foreach (keys %namespaces){

    $stsel->bind_param(1, $_);
    $stsel->execute or die "Couldn't execute SELECT statement: " . $stsel->errstr;

    while ($row = $stsel->fetchrow_hashref){
    
	$stsel1->bind_param(1, $row->{page_id});
	$stsel1->execute or die "Couldn't execute SELECT statement: " . $stsel1->errstr;
	
	while ($subrow = $stsel1->fetchrow_hashref){
		
	    $pages{$subrow->{rev_text_id}} = $row->{page_id};
	}
    }
}

$stsel = $mw_dbh->prepare("SELECT * FROM ${table_prefix}text WHERE old_id = ?") or die "Couldn't prepare SELECT statement: " . $mw_dbh->errstr;

my $stins = $mw_dbh->prepare("INSERT INTO ${table_prefix}html (old_id, old_text, old_flags) VALUES (?, ?, ?)")  or die "Couldn't prepare INSERT statement: " . $mw_dbh->errstr;

my $elem;        

foreach $txtId (keys %pages){

#print "Rev (main): $rev\n";

    $stsel->bind_param(1, $txtId);
    $stsel->execute or die "Couldn't execute SELECT statement: " . $stsel->errstr;


    while ($row = $stsel->fetchrow_hashref()){

	foreach $elem (keys(%{$row})){
    
	    $stins->bind_param(1, $row->{$elem}) if $elem eq 'old_id';
#	    $stins->bind_param(2, text2html($row->{$elem}, $pages{$rev})) if $elem eq 'old_text';
	    $stins->bind_param(2, text2html($row->{$elem})) if $elem eq 'old_text';
	    $stins->bind_param(3, $row->{$elem}) if $elem eq 'old_flags';
	}

	$stins->execute or die "Couldn't execute INSERT statement: " . $stins->errstr;
    }
                    
}            

$mw_dbh->disconnect;
$tiki_dbh->disconnect;



sub listAttachments{

    my @res;
    my @index = ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f');
    my ($listing, $ind);
    
    foreach $ind (@index){
    
	$listing = `ls $mw_att_base/thumb/$ind/*/*/*`;
	push(@res, split(/\n/, $listing)); 
    }	

    foreach $ind (@index){

	$listing = `ls $mw_att_base/$ind/*/*`;
	push(@res, split(/\n/, $listing)); 

    }
    return @res;
}



sub text2html 
{

    my $text = shift;

    # Cut off some internal MediaWiki formatting that we can't convert
    $text =~ s/__NOEDITSECTION__//g;
    $text =~ s/__NOTOC__//g;
    # $text =~ s/\{\{.*}}//gi;


    # Convert includes
    $text =~ s/<noinclude>.*<\/noinclude>//gms;
    $text = convertInclude($text) if $text =~ /\{\{[^}]+}}/;
    
    # Remove block formatting tags and convert some headers inside those tags to wiki markup
    $text =~ s/<\/*div[^>]*>//g;
    $text =~ s/(\[\[.+)<\/*span[^>]*>(.*Welcome.*|.*Wiki.*|.*VS.*|.*Managed.*)<\/span>(?:<\/*span[^>]*>(.*Welcome.*|.*Wiki.*|.*VS.*|.*Managed.*)<\/span>)*(]])/
                            eval{my $tmp1 = $1;
                            my $tmp2 = $2;
                            my $tmp3 = $3;
                            my $tmp4 = $4;
                            $tmp2 =~ s(')[]g;
                            $tmp3 =~ s(')[]g;
                            return "=$tmp1$tmp2$tmp3$tmp4=";}/egi;

    $text =~ s/<\/*span[^>]*>(.*Welcome.*|.*Wiki.*|.*VS.*|.*Managed.*)<\/span>/eval{my $tmp=$1;
                            $tmp =~ s(')[]g; #'
                            return "=$tmp=";}/egi;

    $text =~ s/<\/*span[^>]*>//g;
    $text =~ s/&nbsp;/ /g;
   
    # Process image galleries
    $text =~ s/(<gallery[^>]*>.+?<\/gallery>)/convertGallery($1)/egs;
 

    # Process tables
    $text = convertTables($text) if ( $text =~ /\{\|/m );
    
    print "WARNING!!! Some MW page contains table(s) without closing tag '|}'! Table(s) can't be converted!\nRun Orem_check_tables.pl to get page name.\n\n" if ( $text =~ /\{\|/m );

    # Convert whole page to html
    $text = wikiformat($text, {link => \&_html_link}, {prefix => 'tiki-index.php?page=', prefix_fdnl => 'tiki-download_wiki_attachment.php?attId='});

    # Fix some HTML-related issues produced by converter 
    $text =~ s/<p>(&lt;html.+?\/html&gt;)<\/p>/eval{my $tt = $1;
                                     $tt =~ s(&lt;)[<]g;
                                     $tt =~ s(&gt;)[>]g;
                                     $tt =~ s(&quot;)["]g; #"
                                     $tt =~ s(<\/*p>)[]g;
                                     return $tt;}/egs;

    $text =~ s/&lt;(\/*ul|\/*td|\/*tr|\/*table)&gt;/<$1>/g;
    

    # Remove all new line characters to eliminate extra spaces in tiki
#    $text =~ s/\n//g;
    $text =~ s/(<\/*\w+>)\n+/$1/g;
    # Escape carets to avoid boxes
    $text =~ s/\^/~np~\^~\/np~/g;
    
    return $text;
} 


sub convertGallery{

    my $txt = shift;
    my $res = '<table>';
    my $perrow = 1;
    my ($width, $tmp, @items, $col, $i);

    $txt =~ m/<gallery([^>]*)>/;
    $tmp = $1 if $1;
    $perrow = $1 if $tmp =~ m/perrow="(\d+?)"/;
    $width = $1 if $tmp =~ m/\widths="(\d+px)"/;

    $txt =~ m/<gallery[^>]*>(.+?)<\/gallery>/s;
    $tmp = $1;
    $tmp =~ s/^\n+//;
    @items = split(/\n/, $tmp);


    $i = 0;

    while($i <= $#items){

        $res .= '<tr>';

        for ($col = 1; $col <= $perrow; $col++){

            if ($i <= $#items){

                    $res .= '<td>';
                    $items[$i] =~ m/(Image:[^\|]+?)\|(.+)/;

                    $res .= "\n[[$1|";
                    $res .= "$width|" if $width;
                    $res .= "$2]]<br>$2\n</td>";

            }

            $i++;
        }

        $res .= '</tr>';

    }

    $res .= '</table>';

    return $res;

}



sub convertInclude
{

    my $txt= shift;
    my $res = '';
    my @text = split(/\n/, $txt);
    my $line;
    my $tmp;

    foreach $line (@text){




        if($line =~ /\{\{Bug\|[^}]+}}/){

            $line =~ s(\{\{Bug\|([^}]+)}})<'[http://bugzilla.corp.omniture.com/show_bug.cgi?id='.$1.' Bug '.$1.']'>eg;

        }elsif($line =~ /\{\{[^}]+}}/){

            $line =~ s/\{\{([^}]+)}}/'{INCLUDE(page=>Template:'.eval{
                                                                    my $tmp = $1;
                                                                    $tmp =~ s(\s)*_*g;
                                                                    return $tmp;}.')}{INCLUDE}'/eg;
        }

    $res .= $line ."\n";

    }

return $res;

}


sub _html_link
{
    my ($tag, $opts, $tags) = @_;

    my ($class, $trailing) = ('', '');
    my ($href, $title, $img_name, @img_opts, $iopt);
    my $px = '';
    my $is_image = 0;
    if ($tag =~ /^\[\[Image\:([^|]+)\|*([^|]*)\|*([^|]*)\|*([^|]*)\|*([^|]*)\|*([^|]*)\|*([^|]*)\]\]$/)        
    {

	$img_name = $1 if $1;
	push(@img_opts, $2) if $2;
	push(@img_opts, $3) if $3;
	push(@img_opts, $4) if $4;
	push(@img_opts, $5) if $5;
	push(@img_opts, $6) if $6;
	push(@img_opts, $7) if $7;

	$is_image = 1;    

	if (($img_name =~ /(\.zip)$/i) || ($img_name =~ /(\.txt)$/i) || ($img_name =~ /(\.doc)$/i) || ($img_name =~ /(\.xls)$/i) || ($img_name =~ /(\.pdf)$/i)){
	    $is_image = 0;
	    $title = $img_name;
	}



	foreach $iopt (@img_opts){
	
	    if ($iopt =~ /\d+px/){
	    
		$px = $iopt;
		
	    }elsif ($iopt eq 'left' || $iopt eq 'right' || $iopt eq 'top' || $iopt eq 'bottom' || $iopt eq 'middle'){
	    
		$trailing .= ' align="'."$iopt".'"' if $is_image;
		
	    }elsif ($iopt =~ /alt=(.*)/){
	    	
	    	$trailing .= ' alt="'."$1".'"' if $is_image;

	    }elsif ($iopt =~ /link=.*/){

		next;

	    }elsif ($iopt =~ /page=.*/){

		next;
		
	    }elsif ($iopt eq 'thumb' || $iopt eq 'thumbnail' || $iopt eq 'frame' || $iopt eq 'framed' || $iopt eq 'frameless' || $iopt eq 'baseline'
		     || $iopt eq 'sub' || $iopt eq 'sub' || $iopt eq 'text-top' || $iopt eq 'text-bottom' || $iopt eq 'center'){
		
		next;
	    
	    }else{
	    
		next if $trailing =~ /alt=/;
	    
		$trailing .= ' alt="'."$iopt".'"' if $is_image;
		$title = $iopt if ! $is_image;
	    }	
	
	}



	$href = $opts->{prefix_fdnl} . importFile($img_name, $px);

    }
    elsif ($tag =~ /^\[\[Media\:([^|]+)\|*([^|]*)\]\]$/)
    {
	$img_name = $1;
	$title = $img_name;
	$title = $2 if $2;
	$href = $opts->{prefix_fdnl} . importFile($img_name, $px);
    
    }
    elsif ($tag =~ /^\[\[([^|#]*)(?:(#)([^|]*))?(?:(\|)(.*))?\]\]$/)
    {

        # Wiki link
	my ($link, $s2, $s3, $s4, $s5) = ($1, $2, $3, $4, $5);
	# Remove trailing and leading spaces
        $link =~ s/^\s+//;
        $link =~ s/\s+$//;
        $link =~ s/(:\s)/:/g;
        
        $href = $opts->{prefix} . Text::MediawikiFormat::_escape_uri $opts, $link if $link;
        $href .= $s2 . Text::MediawikiFormat::_escape_uri $opts, $s3 if $s2;


        if ($s4)
        {
            # Title specified explicitly.
            if (length $s5)
            {
                $title = $s5;
            }
            else
            {
                # An empty title asks Mediawiki to strip any parens off the end
                # of the node name.
                $link =~ /^([^(]*)(?:\s*\()?/;
                $title = $1;
            }
        }
        else
        {
            # Title defaults to the node name.
            $title = $link;
        }
        
        $href =~ s/(\%26amp\%3B)/%26/g;
	$href =~ s/(\%20)/_/g;
        
    }
    elsif ($tag =~ /^\[(\S*)(?:(\s+)(.*))?\]$/)
    {
        # URI
        $href = $1;
        if ($2)
        {
            $title = $3;
        }
        else
        {
            $title = ++$opts->{_uri_refs};
        }
        $href =~ s/'/%27/g;
    }
    else
    {
        # Shouldn't be able to get here without either $opts->{absolute_links}
        # or $opts->{implicit_links};
        $tags->{_schema_regex} ||= Text::MediawikiFormat::_make_schema_regex (@{$tags->{schemas}});
        my $s = $tags->{_schema_regex};

        if ($tag =~ /^$s:[$Text::MediawikiFormat::uricCheat][$Text::MediawikiFormat::uric]*$/)
        {
            # absolute link
            $href = $&;
            $trailing = $& if $href =~ s/[$Text::MediawikiFormat::uriCruft]$//;
            $title = $href;
        }
        else
        {
            # StudlyCaps
            $href = $opts->{prefix} . Text::MediawikiFormat::_escape_uri $opts, $tag;
            $title = $tag;
        }
    }

    if ($is_image)
    {

	return "<img src='$href'$trailing>";
    
    }else{

        return "<a$class href='$href'>$title</a>$trailing";
    }
}


sub importFile {

    my $fname = shift;
    $fname =~ s/(\(|\)|\[|\])/\\$1/g;    
    
    my $cfname = $fname;
    $cfname =~ s/\s/_/g;

    my $dpx = shift;
    my ($res, $pname, $hash, $att, $sth, $sth1, $fpath);
    my $src_path = '';
    my $mm = new File::MMagic;
    
    
    # Get page name and namespace

    $sth = $mw_dbh->prepare("SELECT page_title, page_namespace FROM ${table_prefix}page WHERE page_id = ?") or die "Couldn't prepare SELECT statement: " . $mw_dbh->errstr;
    $sth->bind_param(1, $pages{$txtId});
    $sth->execute or die "Couldn't execute SELECT statement: " . $sth->errstr;    
    while ($res = $sth->fetchrow_hashref){
    
	$pname = $res->{page_title};
	$pname = $namespaces{$res->{page_namespace}}.":".$pname if $res->{page_namespace} != 0;
    
    }
    
    
    # Find file in MWiki location (search through contents of @attachments array)
    
    foreach $fpath (@attachments){
    
    
	if ($dpx && $fpath =~ /^.+\/$dpx-$fname$/i){
	
	    $src_path = $fpath;
	    last;

	}elsif($dpx && $fpath =~ /^.+\/$dpx-$cfname$/i){

	    $src_path = $fpath;
	    last;
	
	}elsif($fpath =~ /^.+\/$fname$/i){

	    $src_path = $fpath;
	    last;

	}elsif($fpath =~ /^.+\/$cfname$/i){

	    $src_path = $fpath;
	    last;
	}
    }
    

    # Nothing found?
    if (!$src_path){
    
	print "WARNING!!! File '$fname' referenced on the page '$pname' was not found!\n\n";
	return '';
    
    }
    
    
    $pname =~ s/\s/_/g;
    $hash = md5_hex("$pname$fname");
    
    
    # Copying file
    
    copy("$src_path", "$tiki_att/$hash") or die "Can't copy $src_path to $tiki_att/$hash: $!";
    
    # Get latest attachment ID from tiki DB
    $sth = $tiki_dbh->prepare('SELECT * FROM tiki_wiki_attachments ORDER BY attId DESC LIMIT 1') or die "Couldn't prepare SELECT statement: " . $tiki_dbh->errstr;
    $sth->execute or die "Couldn't execute SELECT statement: " . $sth->errstr;    
        

    $att = 0;

    while ($res = $sth->fetchrow_hashref){
 
	$att = $res->{attId} if $res->{attId} ;
    }
    
    
    $sth = $tiki_dbh->prepare('SELECT * FROM tiki_wiki_attachments WHERE path = ?') or die "Couldn't prepare SELECT statement: " . $tiki_dbh->errstr;
    $sth->bind_param(1, $hash);
    $sth->execute or die "Couldn't execute SELECT statement: " . $sth->errstr;    

    
    $res = $sth->fetchrow_hashref;


	if (defined $res->{attId}){
	
	
	    return $res->{attId};	
	
	}else{

	    $sth1 = $tiki_dbh->prepare('INSERT INTO tiki_wiki_attachments (attId, page, filename, filetype, filesize, user, path, comment, hits) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)') or die "Couldn't prepare INSERT statement: " . $tiki_dbh->errstr;
	    $sth1->bind_param(1, ($att + 1));
	    $sth1->bind_param(2, $pname);
	    $sth1->bind_param(3, $fname);
	    $sth1->bind_param(4, $mm->checktype_filename($src_path));
	    $sth1->bind_param(5, (stat $src_path)[7]);
	    $sth1->bind_param(6, 'Anonymous');
	    $sth1->bind_param(7, $hash);
	    $sth1->bind_param(8, 'Imported from MediaWiki');
	    $sth1->bind_param(9, 0);
	    $sth1->execute or die "Couldn't execute INSERT statement: " . $sth1->errstr;    
	
	    return ($att + 1);
		
	}
    
}

sub convertTables{


    my $txt = shift;
    my $sstr;


    while($txt =~ m/(\{\|.+?\|\})/s){

        $sstr = $1;

        if ($sstr =~ m/^\{\|.+\{\|.+\|\}/s){

            $txt =~ s/\{\|(.+?\|\})/eval{my $tt = convertTables($1); return ('{|'. $tt);}/es;

        }else{

            $txt =~ s/(\{\|.+?\|\})/convertTable($1)/es;

        }
    }

    return $txt;
}

sub convertTable{

    my $tline = shift;
    my $resline = '';
    my (%table, $row, $cell, $tmp, $j);
    my $i = 0; # Row index


    # Parse the wiki table and collect table data in hash

    $tline =~ s/\|\}$//m;
    $tline =~ s/^\s+//;

    foreach $row (split /\|-/, $tline){

        $j = -1; # Cell index

        # Remove leading spaces
        $row =~ s/^\s+//;
        # Remove leading newline characters
        $row =~ s/^\n+//;


        # If this is the first row in the table
        if ($row =~/^\{\|(.*)/){

            $table{attribs} = $1;
            $row =~ s/^\{\|.*//; # Cut it off
            $row =~ s/^\n+//;
        }

        # We got row with its own set of attributes
        if ($row =~ /^([^\|!].*)/){

            $table{rows}[$i]{attribs} = $1;
            $row =~ s/.+//;
            $row =~ s/^\n+//;
        }


        # We got table with caption
        if ($row =~ /^\|\+(.+)/s){

            $tmp = $1;
            $table{rows}[$i]{type} = 'tc';

            if($tmp =~ /(.+)\|(.+)/s){
                $table{rows}[$i]{attribs} = $1;
                $table{rows}[$i]{data} = $2;
            }else{

                $table{rows}[$i]{data} = $tmp;

            }

            $i++;
            next;
        }

        # We got table in format "one row on one line"
        if ($row =~ m/^\|.+(?:\|\|.+)+/){

            $row =~ s/^\|//; # Remove leading pipe

            foreach $cell (split /\|\|/, $row){
                $j++;

                $table{rows}[$i]{cells}[$j]{type} = 'td';

                if ($cell =~/(.*style=.+?|.*class=.+?|.*align=.+?|.*span=.+?|.*color=.+?|.*width=.+?)\|{1}(.*)/){ # <td> with attributes!

                    $table{rows}[$i]{cells}[$j]{attribs} = $1;
                    $table{rows}[$i]{cells}[$j]{data} = $2;

                }else{

                    $table{rows}[$i]{cells}[$j]{data} = $cell;
                }
            }

            $i++;
            next;

        }
        # The same as above but with table headings, not ordinary cells
        if($row =~ /^!.+(?:!!.+)+/){

            $row =~ s/^!//; # Remove leading exclamation sign

            foreach $cell (split /!!/, $row){
                $j++;

                $table{rows}[$i]{cells}[$j]{type} = 'th';
                $table{rows}[$i]{cells}[$j]{data} = $cell;
            }

            $i++;
            next;
        }

        # If we got here, we have table with one cell per line (row consits of multiple lines)

        foreach $cell (split /\n/, $row){

            if ($cell =~ /^\s*!(.*)/){ # Wow, <th>!

                $j++;

                $tmp = $1;
                $table{rows}[$i]{cells}[$j]{type} = 'th';

                if ($tmp =~/(.*style=.+?|.*class=.+?|.*align=.+?|.*span=.+?|.*color=.+?|.*width=.+?)\|(.*)/){ # Wow, <th> with attributes!
                    $table{rows}[$i]{cells}[$j]{attribs} = $1;
                    $table{rows}[$i]{cells}[$j]{data} = "$2\n";

                }else{

                    $table{rows}[$i]{cells}[$j]{data} = "$tmp\n";
                }

            }elsif ($cell =~ /^\s*\|(.*)/){ # <td>

                $tmp = $1;

                if ($cell =~ m/^(\|[^\|]+)*(\|\|[^\|]*)+/){ # Incredible table markup. But we'll convert it anyway.

                    $j++;

                    if ($1){

			$table{rows}[$i]{cells}[$j]{attribs} = 'align="center"';
                        $table{rows}[$i]{cells}[$j]{type} = 'td';
                        $table{rows}[$i]{cells}[$j]{data} = $1;

                        $table{rows}[$i]{cells}[$j]{data} =~ s/^\|\s*//;

                        $j++;
                    }

                    my @sres = $cell =~ m/\|\|([^\|]*)+?/g;

                    foreach $tmp (@sres){

			$table{rows}[$i]{cells}[$j]{attribs} = 'align="center"';
                        $table{rows}[$i]{cells}[$j]{type} = 'td';
                        $table{rows}[$i]{cells}[$j]{data} = ($tmp ? $tmp : ' ');
                        $j++;

                    }
                    $i++;
                    last;

                }



                $j++;

                $table{rows}[$i]{cells}[$j]{type} = 'td';



                if ($tmp =~/(.*style=.+?|.*class=.+?|.*align=.+?|.*span=.+?|.*color=.+?|.*width=.+?)\|(.*)/){ # <td> with attributes!
                    $table{rows}[$i]{cells}[$j]{attribs} = $1;
                    $table{rows}[$i]{cells}[$j]{data} = "$2\n";
                }else{

                    $table{rows}[$i]{cells}[$j]{data} = "$tmp\n";
                }

            }elsif ($cell =~ /^([^\|!].*)/){ # Cell with multiline value

		if ($j == -1){

		    print "WARNING!!! Incorrect table markup detected on some MW page! Some data is lost!\nRun Orem_check_tables.pl to get page name.\n\n";
		    next;
		}

                    $table{rows}[$i]{cells}[$j]{data} .= "$1\n";

            }

        }

        $i++;
    }


    # Now, let's try to build HTML table

    # First, we create opening tag with attribs
    $resline .= '<table';

    $resline .=  " $1" if $table{attribs} =~ /\s*(style=".+?")\s*/;
    $resline .=  " $1" if $table{attribs} =~ /\s*(class=".+?")\s*/;

    $resline .=  " $1" if $table{attribs} =~ /\s*(cellpadding="\d+")/;
    $resline .=  " $1" if $table{attribs} =~ /\s*(cellspacing="\d+")/;
    $resline .=  " $1" if $table{attribs} =~ /\s*(border="\d+")/;
    $resline .=  " $1" if $table{attribs} =~ /\s*(align=".+?")\s*/;
    $resline .=  " $1" if $table{attribs} =~ /\s*(width=".+?")\s*/;
    $resline .= ">\n";


    for ($i = 0; $i <= $#{$table{rows}}; $i++){

        if($table{rows}[$i]{type} eq 'tc'){

            $resline .= '<caption';
            $resline .=  " $1" if $table{rows}[$i]{attribs} =~ /\s*(style=".+?")\s*/;
            $resline .=  " $1" if $table{rows}[$i]{attribs} =~ /\s*(class=".+?")\s*/;
            $resline .=  " $1" if $table{rows}[$i]{attribs} =~ /\s*(align=".+?")\s*/;
            $resline .=  " $1" if $table{rows}[$i]{attribs} =~ /\s*(valign=".+?")\s*/;
            $resline .= '>';

            $table{rows}[$i]{data} =~ s/\n$//;

            $resline .= $table{rows}[$i]{data} . "</caption>\n";
            next;
        }

        next unless ($#{$table{rows}[$i]{cells}} + 1); # Skip empty rows



        $resline .= '<tr';
        $resline .=  " $1" if $table{rows}[$i]{attribs} =~ /\s*(class=".+?")\s*/;
        $resline .=  " $1" if $table{rows}[$i]{attribs} =~ /\s*(style=".+?")\s*/;

        foreach $tmp (split(/ /, $table{rows}[$i]{attribs})){

            $resline .=  " $1" if $tmp =~ /^(align=".+")/;
            $resline .=  " $1" if $tmp =~ /^(valign=".+?")/;
        }

        $resline .= '>';

        for ($j = 0; $j <= $#{$table{rows}[$i]{cells}}; $j++){


                $resline .= "<$table{rows}[$i]{cells}[$j]{type}";
                $resline .=  " $1" if $table{rows}[$i]{cells}[$j]{attribs} =~ /\s*(style=".+?")\s*/;
                $resline .=  " $1" if $table{rows}[$i]{cells}[$j]{attribs} =~ /\s*(class=".+?")\s*/;
                $resline .=  " $1" if $table{rows}[$i]{cells}[$j]{attribs} =~ /\s*(width=".+?")\s*/;
                $resline .=  " $1" if $table{rows}[$i]{cells}[$j]{attribs} =~ /\s*(rowspan=".+?")\s*/;
                $resline .=  " $1" if $table{rows}[$i]{cells}[$j]{attribs} =~ /\s*(colspan=".+?")\s*/;

                foreach $tmp (split(/ /, $table{rows}[$i]{cells}[$j]{attribs})){

                    $resline .=  " $1" if $tmp =~ /^(align=".+")/;
                    $resline .=  " $1" if $tmp =~ /^(valign=".+?")/;
                }

                $resline .= '>';

                $resline .= $table{rows}[$i]{cells}[$j]{data} . "</$table{rows}[$i]{cells}[$j]{type}>";

        }

        $resline .= "</tr>";
    }

    $resline .= "</table>";

    return $resline;
}


